#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <getopt.h>
#include <omp.h>
#include <curses.h>
#include "fire.h"

void usage() {
    fprintf(stderr,"Usage: %s -p <burn-probability>\n",
            program_invocation_short_name
           );
}

struct user_args get_user_args(const int argc,char **argv) {
    int opt;
    struct user_args args;

#ifdef DEBUG
        fprintf(stderr,"Processing user args\n");
#endif
    args.burn_ratio = 0;
    while((opt = getopt(argc,argv,":p:")) != -1) {
        switch(opt) {
            case 'p':
                args.burn_ratio = strtof(optarg,NULL);
                if(args.burn_ratio == HUGE_VAL) {
                    fprintf(stderr,"Received invalid value: %s, %s:%d\n",
                            strerror(errno),__FILE__,__LINE__);
                    usage();
                    abort();
                }
                break;
        }
    }

    if(args.burn_ratio <= 0 || args.burn_ratio > 1) {
        fprintf(stderr,"Supply valid burn ratio!\n");
        usage();
        abort();
    }

#ifdef DEBUG
    fprintf(stderr,"Burn ratio = %f\n",
            args.burn_ratio
           );
#endif

    return args;
}

struct forest *init_forest(
        const struct user_args args,
        struct drand48_data *rand_buffer) {
    struct forest *burning_forest;
    size_t forest_bytes = sizeof(forest)*MAX_COORD;
    double rand_result;
    unsigned int i;
    int up,down,right,left;

#ifdef DEBUG
    fprintf(stderr,"Initializing field of %d cells\n",MAX_COORD);
#endif

    if((burning_forest = malloc(forest_bytes)) == NULL) {
        fprintf(stderr,"Failed to malloc %zd bytes for forest: %s, %s:%d\n",
                forest_bytes,strerror(errno),__FILE__,__LINE__
               );
        return NULL;
    }

    /* Set each cell to be a fresh tree */
#ifdef OPENMP
#pragma omp parallel for
#endif
    for(i=0;i<MAX_COORD;i++) {
        burning_forest[i].tree = GREEN_TREE;

        /* Cache neighbor information */
        up = i-MAX_X;
        down = i+MAX_X;
        left = i-1;
        right = i+1;

        burning_forest[i].neighbors.up
            = up > 0
            ? up
            : OOB;

        burning_forest[i].neighbors.left
            = left > 0
            ? left
            : OOB;

        burning_forest[i].neighbors.down
            = down < MAX_COORD
            ? down
            : OOB;

        burning_forest[i].neighbors.right
            = right < MAX_COORD
            ? right
            : OOB;
    }

    drand48_r(rand_buffer,&rand_result);
    i = (int)(rand_result*MAX_COORD);
#ifdef DEBUG
    fprintf(stderr,"1st burning tree is at %d\n",i);
#endif

    /* Increment by one to start it burning */
    burning_forest[i].tree++;

    return burning_forest;
}

int burn_trees(
        const struct user_args args,
        struct forest *burning_forest,
        struct forest *next_burning_forest,
        struct drand48_data *rand_buffer,
        WINDOW *curses
        ) {
   unsigned int i,t,cur_burn_trees,started_burn_trees; 

   /* The simulation is over when no burning trees remain */
   t = 0;
   do {
       cur_burn_trees = started_burn_trees = 0;
       for(i=0;i<MAX_COORD;i++) {
           burning_forest[i].tree = next_burning_forest[i].tree;
       }
#ifdef OPENMP
#pragma omp parallel for shared(cur_burn_trees,started_burn_trees)
#endif
       for(i=0;i<MAX_COORD;i++) {
           if(burning_forest[i].tree > GREEN_TREE 
                   && burning_forest[i].tree < BLACK_TREE) {
#ifdef DEBUG
                fprintf(stderr,"%u has been burning for %u turns\n",
                        i,burning_forest[i].tree);
#endif

                next_burning_forest[i].tree++;
                cur_burn_trees++;
                /* cur_burn_trees will potentially double count
                 * burning trees when run in OpenMP mode, but will
                 * never under count so can still be used as a loop
                 * control variable
                 */
                started_burn_trees
                    += burn_neighbors(args,burning_forest,next_burning_forest,
                            i,rand_buffer);
           }
       }
#ifdef DEBUG
       fprintf(stderr,
               "T:%u - %u trees burning from last timestep, %u started this time step\n",
               t,cur_burn_trees,started_burn_trees
             );
#endif
       print_forest(next_burning_forest,curses); 
       t++;
   } while((cur_burn_trees+started_burn_trees) > 0);

   return 0;
}

unsigned short int burn_neighbors(
        const struct user_args args,
        struct forest *burning_forest,
        struct forest *next_burning_forest,
        const unsigned int i,
        struct drand48_data *rand_buffer
        ) {
    unsigned short int started_burn_trees = 0;
    double rand_result;

    if(burning_forest[i].neighbors.up != OOB
            && burning_forest[
                burning_forest[i].neighbors.up].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
#ifdef DEBUG
            fprintf(stderr,"Setting fire to %d\n",
                    burning_forest[i].neighbors.up
                   );
#endif
            next_burning_forest[burning_forest[i].neighbors.up].tree++;
            started_burn_trees++;
        }
    }

    if(burning_forest[i].neighbors.down != OOB
            && burning_forest[
                burning_forest[i].neighbors.down].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
#ifdef DEBUG
            fprintf(stderr,"Setting fire to %d\n",
                    burning_forest[i].neighbors.down
                   );
#endif
            next_burning_forest[burning_forest[i].neighbors.down].tree++;
            started_burn_trees++;
        }
    }

    if(burning_forest[i].neighbors.left != OOB
            && burning_forest[
                burning_forest[i].neighbors.left].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
#ifdef DEBUG
            fprintf(stderr,"Setting fire to %d\n",
                    burning_forest[i].neighbors.left
                   );
#endif
            next_burning_forest[burning_forest[i].neighbors.left].tree++;
            started_burn_trees++;
        }
    }

    if(burning_forest[i].neighbors.right != OOB
            && burning_forest[
                burning_forest[i].neighbors.right].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
#ifdef DEBUG
            fprintf(stderr,"Setting fire to %d\n",
                    burning_forest[i].neighbors.right
                   );
#endif
            next_burning_forest[burning_forest[i].neighbors.right].tree++;
            started_burn_trees++;
        }
    }

    return started_burn_trees;
}

unsigned int calc_2d21d(
        const unsigned int dim,
        const unsigned int x,
        const unsigned int y
        ) {
    return (dim*x)+y;
}

void print_forest(
        const struct forest* burning_forest,
        WINDOW *curses
        ) {
    unsigned int x,y,tree;

    for(x=0;x<MAX_X;x++) {
        for(y=0;y<MAX_Y;y++) {
            wmove(curses,y,x);
            waddch(curses,'0'+burning_forest[calc_2d21d(MAX_X,x,y)].tree);
            /*
             * tree = burning_forest[calc_2d21d(MAX_X,x,y)].tree;
             *
             * 
             * if(tree == GREEN_TREE) {
             *     waddch(curses,COLOR_GREEN);
             * }
             * else if(tree < BLACK_TREE) {
             *     waddch(curses,COLOR_RED);
             * }
             * else {
             *     waddch(curses,COLOR_BLACK);
             * }
             */
        }
    }

    refresh();

    sleep(1);
}
